import inspect, os, copy, json, re
from inspect import Parameter
from print_ext import print, Table, pretty, Line, Printer
from .exceptions import UsageError

INFINITY = 1000000000000

def sig_kinds(fn):
    kinds = set()
    for k in ['generatorfunction']:#'module', 'class', 'method', 'function', , 'generator', 'coroutinefunction', 'coroutine', 'awaitable', 'asyncgenfunction', 'asyncgen', 'traceback', 'frame', 'code', 'builtin', 'methodwrapper', 'routine', 'abstract', 'methoddescriptor', 'datadescriptor', 'getsetdescriptor', 'memberdescriptor']:
        try:
            if getattr(inspect, 'is'+k)(fn): kinds.add(k)
        except:
            pass
    return kinds


def func_name(fn, attr='__name__'):
    try:
        return getattr(fn.__func__, attr)
    except:
        return getattr(fn, attr)


def name_split(name):
    if name.startswith('_'): return
    for a in name.rsplit('__'):
        yield (a[:-1] if a[-1] == '_' else a).replace('_','-')



def dashed(name):
    return ('--' if len(name) > 1 else '-') + name


class ArgType():
    def __init__(self, val):
        self.repeated = False
        if isinstance(val, list):
            self.repeated = True
            val = val[0] if val else ''
        if not callable(val):
            val = type(val)
        self.coerce = str if val==type(None) or val==Parameter.empty else val
        self.is_flag = self.coerce == bool
        if self.coerce == dict: self.coerce = json.loads
        if self.repeated and self.is_flag: raise ValueError("We can't handle repeated bool arguments")


    def __repr__(self):
        return f'[{self.coerce.__name__}]' if self.repeated else self.coerce.__name__


    def merge(self, val, pval):
        if val == None: # append nothing to the list (but make sure it exists)
            return [] if pval == Parameter.empty else pval
        try:
            assert(not (self.is_flag and not isinstance(val, bool))) # Only True can be set on bools
            assert(not (self.coerce==str and not isinstance(val, str))) # Don't coerce None or True to a string
            val = self.coerce(val)
        except Exception as e:
            raise ValueError(Line(f"Couldn't coerce '\b2 {val}\b ' to '\b2 {self}\b '."))
        if self.repeated:
            return ([] if pval == Parameter.empty else pval) + [val]
        elif self.is_flag:
            return val if pval == Parameter.empty else int(pval)+1
        elif self.coerce == json.loads:
            if isinstance(pval, dict) and isinstance(val, dict):
                pval.update(val)
                return pval
            elif isinstance(pval, list) and isinstance(val, list):
                return pval + val
            return val
        else:
            return val
            


class ArgParam(dict):
    def __repr__(self):
        return f"{self.name}:{','.join(self.aliases)}@{self.index}<{self.type}>{self.value}"


    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)

    
    def set(self, val, index):
        try:
            self['value'] = self.type.merge(val, self.value)
        except ValueError as e:
            if isinstance(index, int):
                self.spec.error('TYPE_MISMATCH', f"Type mismatch on \b1 {ordinal(index)}\b  parameter.  {e}")
            else:
                self.spec.error('TYPE_MISMATCH', f"Parameter '\b1 {index}\b ' type mismatch.  {e}")


    @property
    def ordinal(self):
        n = self.index
        if not n: return ''
        if n%10==1 and n!=11: return f"{n}st"
        if n%10==2 and n!=12: return f"{n}nd"
        if n%10==3 and n!=13: return f"{n}rd"
        return f"{n}th"



    @staticmethod
    def new(spec, /, **kwargs):
        args = dict(index=0, aliases=[], type=None, default=Parameter.empty, kind=Parameter.KEYWORD_ONLY)
        args.update(kwargs)
        args['spec'] = spec
        args['value'] = Parameter.empty
        param = ArgParam(**args)
        spec.params[param.name] = param
        if param.aliases:
            aliases = set(param.aliases)
            aliases.add(param.name)
            aliases.update([a.replace('-','_') for a in aliases])
            for a in aliases: spec.set_alias(a, param)
        return param




class ArgSpec():
    kw_re = re.compile(r'--?[a-zA-Z][-\w]*(#([1-9][0-9]*)?)?$')

    def __init__(self, fn, is_method=False):
        self.errors = []
        self.unknown = []
        self.alias = {}
        self.params = {}
        self.fn = fn
        self.kinds = set()
        self.args = None
        self.kwargs = None
        self.has_self = False
        # Are we cloning?
        if isinstance(fn, ArgSpec):
            for param in fn.params.values():
                ArgParam.new(self, **param)
            self.has_self, self.kinds, self.fn, self.sig = fn.has_self, fn.kinds, fn.fn, fn.sig
            return
        # We are not a clone
        try:
            self.sig = inspect.signature(self.fn)
        except:
            self.sig = None
            return
        # Figure out a set of kinds
        self.kinds.update(sig_kinds(fn))
        for i, name in enumerate(self.sig.parameters):
            p = self.sig.parameters[name]
            if p.kind == Parameter.VAR_POSITIONAL:
                self.kinds.add('*')
                continue
            if p.kind == Parameter.VAR_KEYWORD:
                self.kinds.add('**')
                continue

            if name.startswith('_') and p.kind == Parameter.POSITIONAL_ONLY:
                raise UsageError(fn, f"Private parameter '{name}' must be a keyword parameter.")
            ArgParam.new(self,
                name = name,
                type = ArgType(p.default if p.annotation == Parameter.empty else p.annotation),
                aliases = list(name_split(name)) if p.kind in [Parameter.POSITIONAL_OR_KEYWORD, Parameter.KEYWORD_ONLY] else [],
                default = p.default,
                index = 0 if p.kind==Parameter.KEYWORD_ONLY else i+1,
                kind = p.kind,
            )
        # The help parameter is always available
        ArgParam.new(self, name='', type=ArgType(bool), default=False, aliases=['help', 'h'])
        self.has_self = 'self' in self.params and self.params['self'].index == 1


    def __bool__(self):
        return self.sig != None


    @property
    def retval(self):
        return Parameter.empty if self.sig==None else self.sig.return_annotation


    def __call__(self, argv):
        spec = ArgSpec(self)
        spec.arg_bind(argv)
        return spec
        

    def error(self, code, *args, **kwargs):
        self.errors.append((code, Line(*args, **kwargs)))


    def set_alias(self, a, param):
        if a == 'self' and (param.index != 1 or param.name != 'self'):
            raise UsageError(self.fn, f"Invalid parameter \b1 {param.name}\b .\v\b1 self\b  is reserved.")
        if a in self.alias:
            prev = self.alias[a]
            if a in ['h','help']:
                raise UsageError(self.fn, f"Invalid parameter \b1 {prev.name}\b .  The alias \b1 {a}\b  is reserved for the help flag.")
            raise UsageError(self.fn, f"Argument \b1 {a}\b  was defined multiple times. \b2 {prev.name} {param.name} ")  
        self.alias[a] = param


    def arg_bind(self, argv):
        self.argv = argv
        # Parse positional args
        for param in self.params.values():
            if not param.index: break
            while self._parse_parg(param, argv) and param.type.repeated: pass
        # Parse keyword args
        while self._parse_kwargs(argv): pass
        # Set default values
        for name, param in self.params.items():
            if param.value != Parameter.empty: continue
            #param['value'] = copy.deepcopy(param.default) # bypass the coerce
            if param.default != Parameter.empty: continue
            # This parameter remains unset!
            if param.name.startswith('_') or (self.has_self and param.index==1):
                continue # self or hidden parameters will be supplied at runtime
            elif not param.aliases and param.ordinal:
                self.error('NO_VALUE', f"No value supplied for the \b1 {param.ordinal}\b  positional parameter (\b2 {name}\b )")
            else:
                aka = ' (\b1 '+ ' '.join(dashed(a) for a in param.aliases) + '\b )'
                pos = f'\b1 {param.ordinal}\b  positional' if param.ordinal else 'keyword-only'
                self.error('NO_VALUE', f"No value supplied for the {pos} parameter{aka}.")
        # De-list unknowns
        for param in self.unknown:
            if not isinstance(param.value, list): continue
            if len(param.value) > 1: continue
            param['value'] = param.value[0]
            param.type.repeated = False
        # Unknown parameters
        if '**' not in self.kinds and self.unknown:
            for param in sorted(self.unknown, key=lambda x: x.name):
                self.error('UNK_PARAM', f"Unknown parameter: \b1 {dashed(param.name)}\b2  {param.value!r}")
        # Build args, kwargs
        self.args, self.kwargs = [], {}
        for p in self.params.values():
            if p.index:
                self.args.append(p.value)
            elif p.value != Parameter.empty:
                self.kwargs[p.name] = p.value
        if '*' in self.kinds:
            self.args += list(argv)
            argv[:] = []


    def _parse_kwargs(self, argv):
        ''' Try to take a keyword argument from argv
        '''
        if not argv: return
        if argv[0][0] != '-': return # A potential command name
        if len(argv[0]) == 1: return argv.pop(0) and False # End of kwargs
        if not ArgSpec.kw_re.match(argv[0]):
            return self.error('BAD_KW', f"Invalid keyword argument \b1 {argv[0]}\b.")
        arg = argv.pop(0)
        dashes = 1 + int(arg[1] == '-')
        arg = arg.rsplit('#',1)
        arg, count = arg[0][dashes:], (arg[1] if len(arg) == 2 else False)
        ks = [arg] if dashes==2 else list(arg)
        for i, k in enumerate(ks):
            if k == 'self':
                self.error('SELF', f"\b1 self\b  can not be set from the command line.  It must be set from the parent's return value.")
                continue
            last = i == len(ks)-1
            if not self._parse_kwarg(k, '-'*dashes + k, last, argv, last and count): return # abort
        return True


    def _parse_kwarg(self, k, kdash, last, argv, count):
        if count != False: kdash += f'#{count}'
        if k in self.alias:
            param = self.alias[k]
        else: # Unknown parameter
            param = ArgParam.new(self, name=k, aliases=[k], type=None if last else ArgType(bool))
            self.unknown.append(param)
        # Are we trying to use -k#3 with a non-repeated value?
        if count != False and param.type and not param.type.repeated:
            return self.error('NOT_LIST', f"Invalid array# parameter \b1 {kdash}\b for a non-list type \b2 {param.type}\b .")
        # Turn count into a valid number of items to get
        try:
            need = 1 if count==False else int(count) if count else INFINITY
            assert(need > 0)
        except:
            self.error('BAD_LIST', f"Invalid array# parameter: \b1 {kdash}\b .")
        # If we know that we are a flag then don't bother to grab one
        if not last or param.type and param.type.is_flag:
            param.set(True, kdash)
            return True
        # We probably need a value so grab one
        val = self._grab_one(argv)
        if param.type == None:
            # We know the unknown parameter type now
            if val == None:
                param.type = ArgType(bool)
                param.set(True, kdash)
                return True
            param.type = ArgType([])
        # Do we only need a single value?
        if count == False:
            if val == None: return self.error('KW_VAL_MISSING', f"Keyword parameter \b1 {kdash}\b  is missing a value.")
            param.set(val, kdash)
            return True
        # We want `need` values
        got = 0
        while True: # Grab multiple values for our array
            if val == None:
                if need == INFINITY: # --key# terminated
                    param.set(None, kdash)
                    break 
                # Didn't get enough values
                return self.error('LIST_TOO_FEW', f"Expected \b1 {need}\b  values but only received \b1 {got}\b .")
            param.set(val, kdash)
            got += 1
            if got == need: break
            val = self._grab_one(argv)
        return True


    def _parse_parg(self, param, argv):
        ''' Try to take a positional argument from argv
        '''
        if param.name.startswith('_'): return
        v = self._grab_one(argv)
        if v == None: return
        param.set(v, param.index)
        return True


    def _grab_one(self, argv):
        if not argv: return
        arg = argv.pop(0)
        if arg[0] == '-':
            l = len(arg)
            if l == 1: return # This indicates that arguments are done
            if arg[1] not in '0123456789.': # Not negative number
                argv.insert(0, arg[1:] if arg == '-'*l else arg)
                return # A keyword parameter
            # Negative number
        if arg.startswith('\\'): arg = arg[1:]
        return arg


    def __pretty__(self, *, _depth, depth, **kwargs):
        p = Printer()
        p(f"\b1 {func_name(self.fn,'__qualname__')}[\b2 {' '.join(self.kinds)}\b ]\b3 {self.args or '()'}{self.kwargs or '{}'}")
        tbl = Table(0,0,0,0,0)
        tbl.cell("C0", just='>')
        tbl.cell("C1", style='1', just='>')
        tbl.cell("R0", style='w!')
        tbl('Aliases\tName\tPos\tType\tDefault\t')
        for v in self.params.values():
            if v.name=='': continue # Skip help
            tbl(' '.join(sorted(v.aliases)) if v.aliases else ' ', '\t')
            tbl(v.name, '\t')
            tbl(v.index or ' ', '\t')
            tbl(v.type, '\t')
            tbl(pretty('\br *req*' if v.default == Parameter.empty else v.default, _depth=_depth+1, depth=depth-1, **kwargs), '\t')
        p(tbl)
        for e in self.errors:
            p(e[1])
        return p



import pytest
from yaclipy import Command, sub_cmds
from yaclipy.exceptions import CmdError
from print_ext import Printer, Line
from .testutil import printer


def test_CmdError_pretty():
    def bob(jump__j):
        '''bobby'''

    def x(razor__r):
        '''About time.'''

    def cob(*args):
        pass

    @sub_cmds(bob_=bob, a_long_name=x)
    def f():
        ''' Short description.
        Log description starts here



        '''

    cmd = Command(f)
    err = CmdError(cmd=cmd, errors=[Line('bad')])
    o,p = printer(color=False, ascii=True)
    p.pretty(err)
    print(o.getvalue())
    assert(o.getvalue() == '''
 * a-long-name About time.

 * bob         bobby

---

Short description.

Log description starts here

-[ CmdError ]-

 * bad
''')


@pytest.mark.xfail(reason='Not implemented')
def test_doc_testing():
    assert(False)
    


import pytest
from yaclipy import sub_cmds
from yaclipy.exceptions import CmdError
from .testutil import exe


def foo(say, times__t=1) -> str:
    ''' Say something multiple times

    Parameters:
        <message>, --say <message>
            What you want to say
        <int>, --times <int>, -t <int>
            How many times you want to say it.
    '''
    return ' '.join([say] * times__t)


def test_intro_0():
    assert(exe(foo,'-t 3 --say Ho') == 'Ho Ho Ho')


def test_intro_1():
    assert(exe(foo, ['Hello World', '2']) == 'Hello World Hello World')


def test_intro_2():
    assert(exe(foo, 'go --times 3 upper') == 'GO GO GO')


def test_intro_3():
    assert(exe(foo, '\\--times') == '--times')


def test_posvkw():
    def f(a=3, /, banana__b='hi', *, carrot__c:int=None): return a, banana__b, carrot__c
    assert(exe(f, '4 bye --carrot 42') == (4, 'bye', 42))
    assert(exe(f, '4 -c 42 -b bye') == (4, 'bye', 42))


def test_flags():
    def f(*, verbose__v=False, times__t:int): return verbose__v, times__t
    assert(exe(f, '-vt 3 --verbose') == (2,3))
    assert(exe(f, '-vv --times 3') == (2,3))

def test_names():
    def f(*, if_=1, happy_days=2, lots__of__aliases__t__q=3, _hidden=4): return if_,happy_days,lots__of__aliases__t__q, _hidden
    assert(exe(f, '--if 10 --happy-days 20 --happy_days 200 --lots 30 --of 40 --aliases 50 -t 60 -q 70') == (10, 200, 70, 4))
    

def test_subcmds():

    def foo(*, name, _value): return 'foo', name, _value

    def bar(*, name, _value): return 'bar', name, _value

    @sub_cmds(foo, baz=bar)
    def root(*, verbose__v=False):
        return dict(name='jim', _value = 'hi' * int(verbose__v))

    with pytest.raises(CmdError):
        exe(root, '-v foo -h') == ('foo', 'jim', 'hi')
    assert(exe(root, '-vv baz --name bob') == ('bar', 'bob', 'hihi'))


def test_readme_generator(capfd):
    def show(*, _input):
        x, xxx = _input
        print(f'3^{x} == {xxx}')

    @sub_cmds(show)
    def foo(*, times__t=3):
        for i in range(times__t):
            yield i, pow(3,i)

    exe(foo, '-t 4 show')
    out, _ = capfd.readouterr()  
    assert(out == '3^0 == 1\n3^1 == 3\n3^2 == 9\n3^3 == 27\n')


def test_readme_lists():
    def f(a:int, b:[float], c=[]): return a,b,c
    assert(exe(f, '3 1.1 -.1 1e3 - 66 \\-apples') == (3, [1.1, -0.1, 1e3], ['66', '-apples']))
    assert(exe(f, '-c 66 -c \\-apples -b#3 1.1 -0.1 1e3 -a 3') == (3, [1.1, -0.1, 1e3], ['66', '-apples']))
    assert(exe(f, '3 1.1 - -c# 66 \\-apples - -b#2 -.1 1e3') == (3, [1.1, -0.1, 1e3], ['66', '-apples']))


def test_readme_json():
    def f(*, x={}, y:dict): return x, y
    assert(exe(f, ['-x','{"x":[1,2,3]}','-y','null']) == ({'x':[1,2,3]}, None))
    

def test_readme_argv():
    def f(first=None, *files, verbose__v=False): return first, files, verbose__v
    assert(exe(f, 'a b c') == ('a', ('b','c'), False))
    assert(exe(f, '- a b c') == (None, ('a', 'b','c'), False))
    assert(exe(f, '- - -v b c') == (None, ('-v', 'b','c'), False))
    assert(exe(f, '-- -v b c') == (None, ('-v', 'b','c'), False))


def test_readme_kwargs():
    def f(a=False, **kwargs) -> str:
        a == True
        kwargs == {'x':True, 'd':['33','44'], 'apple':'x', 'banana':True}
        return f'{a} {kwargs}'
    assert(exe(f, '-axd 33 -d 44 --apple x --banana - upper') == "TRUE {'X': TRUE, 'D': ['33', '44'], 'APPLE': 'X', 'BANANA': TRUE}")
        


# SPDX-FileCopyrightText: 2023-present Aaron <aaron@framelunch.jp>
#
# SPDX-License-Identifier: MIT
from .bootstrap import ensure_requirements
from .command import Command, sub_cmds
from .config import Config



import io
from print_ext import print, Flattener, PrettyException
from yaclipy import Command, sub_cmds
from yaclipy.arg_spec import ArgSpec

DEBUG = False # Show more verbose info

def _to_args(v):
    if not v: return []
    if isinstance(v, str): return v.split(' ')
    return v



def bind(fn, args):
    try:
        cmd = Command(fn)
        cmd(_to_args(args))
        if DEBUG:
            print('--------- cmd.run_spec -----------')
            print.pretty(cmd.run_spec)
            print.pretty(cmd.run_spec.argv)
        return cmd.run_spec.args, [(k,cmd.run_spec.kwargs[k]) for k in sorted(cmd.run_spec.kwargs)]
    except PrettyException as e:
        if DEBUG: print.pretty(e)
        raise e
    


def bind_unused(fn, args):
    try:
        cmd = Command(fn)
        cmd(_to_args(args))
        if DEBUG:
            print('--------- cmd.run_spec -----------')
            print.pretty(cmd.run_spec)
            print.pretty(cmd.run_spec.argv)
        return []
    except PrettyException as e:
        if DEBUG: print.pretty(e)
        return e.cmd.run_spec.argv



def bind_err(fn, args):
    try:
        cmd = Command(fn)(_to_args(args))
        if DEBUG:
            print('--------- cmd.run_spec -----------')
            print.pretty(cmd.run_spec)
            print.pretty(cmd.run_spec.argv)
        return set()
    except PrettyException as e:
        if DEBUG: print.pretty(e)
        errs = {}
        if hasattr(e, 'cmd'):
            for e in e.cmd.run_spec.errors:
                errs[e[0]] = errs.get(e[0],0) + 1
        errs = set([k + (f'x{v}' if v > 1 else '') for k,v in errs.items()])
        errs.discard('HELP')
        return errs



def exe(fn, args, **incoming):
    try:
        cmd = Command(fn)(_to_args(args))
        return cmd.run(incoming)        
    except PrettyException as e:
        if DEBUG: print.pretty(e)
        raise e


def printer(**kwargs):
    o = io.StringIO()
    p = Flattener(stream=o, **kwargs)
    return o,p
   

def tostr(*args, **kwargs):
    o,p = printer(**kwargs)
    p(*args, **kwargs)
    return o.getvalue()

    


#!/usr/bin/env python
import sys, os
from print_ext import print, PrettyException, Table
import yaclipy as CLI


def main(*files, _input, upper__u=False):
    ''' List files

    Parameters:
        --upper, -u
            Make the names all uppercase
    '''
    tbl = Table(0,1,tmpl='kv')
    for f in files:
        name,ext = os.path.splitext(f)
        tbl(name.upper() if upper__u else name, '\t', ext, '\t')
    print(tbl, f'\v--- \b2 {_input}\b  {len(files)} files ---')


if __name__ == '__main__':
    try:
        CLI.Command(main)(sys.argv[1:]).run(os.path.basename(sys.argv[0]))
    except PrettyException as e:
        print.pretty(e, pad=1)



import sys, os, shutil
from print_ext import print, PrettyException, Text
from subprocess import run


class EnvSetup(PrettyException): pass


def ensure_requirements(req, venv):
    req_lock = os.path.splitext(req)[0] + '.lock'
    cur_lock = os.path.join(venv, os.path.basename(req_lock))
    if run(['diff', req_lock, cur_lock]).returncode == 0: return
    print('\bwarn Requirements changed.')
    if run(['python', '-m', 'pip', 'install', '--require-virtualenv', '--compile', '-U', 'pip']).returncode: # yaclipy
        raise EnvSetup(msg='Pip upgrade failed.')
    if not os.path.exists(req_lock):
        print('\vInstalling requirements rom \b2$', req)
        if run(['python', '-m', 'pip', 'install', '--require-virtualenv', '--compile', '-r', req]).returncode:
            raise EnvSetup(msg='Failed to install some requirements.')
        print('\vFreezing to \b2$', req_lock)
        if run(f'python -m pip freeze --local > "{req_lock}"', shell=True).returncode:
            raise EnvSetup(msg="Pip freeze failed")
    else:
        if run(['python', '-m', 'pip', 'install', '--require-virtualenv', '--compile', '-r', req_lock]).returncode:
            raise EnvSetup(msg='Failed to install some requirements.')
    shutil.copy(req_lock, cur_lock)



from yaclipy import Config

prefix = Config.var('A prefix to distinguish configurations.', '')
title = Config.var('The app title', 'Awesome App')


@Config.option()
def dev():
    prefix('dev')


@Config.option()
def prod():
    prefix('prod')


@Config.option()
def test():
    prefix('test')


with Config.include: import local.config



#!/usr/bin/env python
import os, sys
from subprocess import call
from os.path import join, abspath, split, exists

VENV_DIR='.python' # This can be moved if needed
REQ = 'pyutil/requirements.txt' # This can be moved/renamed

def abort(*reason):
    print('\n'+'!'*75,'\nERROR:', *reason)
    print('Look above for the specific error.','\n'+'!'*75,'\n')
    import shutil
    shutil.rmtree(VENV_DIR, ignore_errors=True)
    sys.exit(1)

os.chdir(split(abspath(__file__))[0]) # Make the cwd the same as this file
if sys.prefix == sys.base_prefix: # Not in the virtual env
    new = not exists(VENV_DIR)
    if new and call(['python3', '-m','venv',VENV_DIR]):
        abort("Couldn't create python3 virtual environment at", VENV_DIR)
    os.environ['PATH'] = join(VENV_DIR,'bin') + os.pathsep + os.environ['PATH']
    if new and call(['python', '-m', 'pip', 'install', '-e' '../..']):
        abort("Couldn't install yaclipy into the virtual environment")
    os.execvp('python', ['python', './cli.py'] + sys.argv[1:])

# Now we are running in the virtual environment.
# Install REQ requirements and run the main command.
import yaclipy as CLI
from print_ext import print, PrettyException

try:
    CLI.ensure_requirements(req=REQ, venv=VENV_DIR)
    from pyutil.main import main
    CLI.Command(main)(sys.argv[1:]).run()
except PrettyException as e:
    print.pretty(e, pad=1)



import re
from inspect import Parameter
from print_ext import print
from yaclipy.arg_spec import ArgSpec
from .testutil import bind, bind_err, bind_unused


def _tsterr(spec, args, *errs):
    matcherrs = list(errs)
    matched = []
    s = spec({}, args.split(' ') if args else [])
    if not s.errors: raise Exception("No errors thrown")
    errs = list(errs)
    def _match(e):
        if not (errs and re.match(matcherrs[0], str(e).lower())): return
        matched.append(matcherrs.pop(0))
        return True
    errors = [e for e in s.errors[:-1] if not _match(e)]
    for e in errors:
        print(e)
    assert(matched == errs)



def test_bind_positional():
    def f(a, b, /): pass
    assert(bind(f, '\\ -33') == (['','-33'], []))
    assert(bind(f, '-.1 -0') == (['-.1','-0'], []))



def test_bind_pargs():
    def f(a, *args): pass
    assert(bind(f, 'a - b c -- d') == (['a','b','c','--','d'], []))



def test_bind_pos_slists():
    def f(a:[], b, c=[2], *args): pass
    assert(bind(f, 'a b - c -- -k -') == ([['a','b'],'c', Parameter.empty, '-k','-'], []))



def test_bind_parg_lacking():
    def f(*, a=1, b=2): pass
    assert(bind(f, '-') == ([],[]))
    


def test_bind_parg_use_defaults():
    def f(a=3, b=4): pass
    assert(bind(f, '- -4 -a 1') == ([1, -4], []))



def test_bind_kwargs_bools():
    def f(*, test__x=False, y=False): pass
    assert(bind(f, '-xxy --test --y -yyy') == ([], [('test__x',3), ('y',5)]))



def test_bind_kwargs_kwargs():
    def f(*, x, y=9, **kwargs): pass
    assert(bind(f, '-x \\ --and the --and bob -bt jim -bbz -') == ([], [('and', ['the','bob']), ('b',3), ('t','jim'), ('x',''), ('z',True)]))



def test_bind_triple_dash():
    def f(a=1, b=2, c=3, d=[0], *args): pass
    assert(bind(f, '--- 5 6 - 7') == ([Parameter.empty,Parameter.empty,Parameter.empty,[5,6],'7'], []))



def test_bind_kwargs_bool_val_mix():
    def f(*, x, y=9, z=False): pass
    assert(bind(f, '-zx 3') == ([], [('x','3'), ('z', True)]))
    assert(bind_err(f, '-xz 3') == {'NO_VALUE', 'UNUSED', 'TYPE_MISMATCH'})



def test_bind_kwargs_ill_formed():
    def f(*, x, y=9, z=False): pass
    assert(bind_err(f, '--x 3 ---y 3') == {'BAD_KW', 'UNUSED'})



def test_bind_kwargs_overwrite():
    def f(*, x, y=9, z=False): pass
    assert(bind(f, '-x 3 -x 99') == ([], [('x','99')]))



def test_parse_kwargs_unknown():
    def f(*, x, y=9, z=False): pass
    assert(bind_err(f, '--x 3 -ztt --cats') == {'UNK_PARAMx2'})
    


def test_bind_KW_VAL_MISSING():
    def f(*, x, y=9, z=False): pass
    assert(bind_err(f, '-zz -x - --y') == {'NO_VALUE', 'UNUSED','KW_VAL_MISSING'})



def test_bind_kwargs_done():
    def f(*, x, y=9, z=False): pass
    assert(bind_unused(f, '-x 3 -.3 -y 10') == ['-.3','-y','10'])
    assert(bind_unused(f, '-x 3 done -y 10') == ['done', '-y', '10'])
    assert(bind_unused(f, '-x 3 - -y 10') == ['-y','10'])
  


def test_bind_defaults_bypass_coerce():
    def f(*, x:[int]=34): pass
    assert(bind(f, '') == ([], []))



def test_bind_no_value():
    def f(def___if_, *, _hidden, bob__x, y=9, z=False): pass
    assert(bind_err(f, '-zz -y 3') == {'NO_VALUEx2'})



def test_bind_unknown_missing():
    def f(**kwargs): pass
    assert(bind_err(f, '--bob hi --bob') == {'KW_VAL_MISSING'})
    assert(bind(f, '--bob') == ([],[('bob',True)]))
    assert(bind(f, '--bob --bob') == ([], [('bob',2)]))
    assert(bind(f, '-ax -a -a') == ([], [('a',3), ('x',True)]))
    assert(bind(f, '--bob hi') == ([], [('bob','hi')]))
 


def test_bind_kw_array():
    def f(*, flag__f=False, a:[float], b:[]=[1,2,3]): pass
    assert(bind(f, '-a 1 -b 2') == ([], [('a',[1]), ('b',['2'])]))
    assert(bind(f, '-a 1 --a 2 -b#1 2') == ([],[('a',[1, 2]), ('b',['2'])]))
    assert(bind(f, '-fb#3 1 2 \\- -ffa# 2 -.3 8 3 -1') == ([],[('a',[2,-0.3,8,3,-1]), ('b',['1','2','-']), ('flag__f',3)]))



def test_bind_kw_bad_array():
    def f(*, flag__f=False, a:[float]): pass
    assert(bind(f, '-a 1 --a#3 1 1 1') == ([], [('a',[1,1,1,1])]))
    assert(bind_err(f, '-a 1 -f#3 1 1 1') == {'NOT_LIST', 'UNUSED'})



def test_bind_LIST_TOO_FEW():
    def f(a:[int]): pass
    assert(bind_err(f, '-a#4 1 2') == {'LIST_TOO_FEW'})
    assert(bind_err(f, '-a 4 -a#1') == {'LIST_TOO_FEW'})



def test_bind_pos_array():
    def f(a:[], *, t=None): pass
    assert(bind(f, 'a b -t c') == ([['a','b']], [('t','c')]))



def test_bind_args_and_kwargs():
    def f(*args, **kwargs): pass
    assert(bind(f, 'a --bob c d e - f') == (['a','--bob','c','d','e','-','f'], []))
    assert(bind(f, '-ax --bob - -x d - f') == (['d','-','f'], [('a',True),('bob',True), ('x',2)]))
    

def test_bind_self():
    def f(self, my__m='3', **kwargs): pass
    assert(bind_err(f, '-m happy --self bad') == {'SELF', 'UNUSED'})


def test_bind_empty_array():
    def f(a:[]): pass
    assert(bind(f, '-a# -') == ([[]], []))



import pytest
from inspect import Parameter
from yaclipy.arg_spec import ArgType



def test_ArgType_str():
    assert(str(ArgType(3)) == 'int')
    assert(str(ArgType(int)) == 'int')
    assert(str(ArgType(None)) == 'str')
    assert(str(ArgType(Parameter.empty)) == 'str')
    assert(str(ArgType(3.1)) == 'float')
    assert(str(ArgType([])) == '[str]')
    assert(str(ArgType([int])) == '[int]')
    assert(str(ArgType([3.3])) == '[float]')



def test_argtype_coerce():
    assert(ArgType(3).coerce('-33') == -33)
    assert(ArgType(int).coerce('-33') == -33)
    assert(ArgType(None).coerce('-33') == '-33')
    assert(ArgType(3.1).coerce('-3.3') == -3.3)



def test_argtype_merge_bool_to_str():
    a = ArgType(None)
    with pytest.raises(ValueError):
        a.merge(True, Parameter.empty)



def test_argtype_merge_list_str():
    a = ArgType([])
    assert(a.merge('a', Parameter.empty) == ['a'])
    assert(a.merge('a', ['b']) == ['b', 'a'])
    with pytest.raises(ValueError):
        assert(a.merge(3, ['1','2']) == ['1','2','3'])
    assert(a.merge('3', ['1','2']) == ['1','2','3'])



def test_argtype_merge_bool():
    a = ArgType(True)
    assert(a.merge(False, Parameter.empty) == False)
    assert(a.merge(True, False) == 1)
    assert(a.merge(True, True) == 2)
    with pytest.raises(ValueError):
        print(a.merge('x', 1))



def test_argtype_merge_flaot():
    a = ArgType([3.14])
    assert(a.merge("3.1", [2.1]) == [2.1, 3.1])



def test_argtype_dict():
    a = ArgType({'k':3})
    print(a)
    assert(a.merge('{"a":[1,2,3]}', {'b':[6]}).keys() == {'a','b'})
    assert(a.merge('33.3', {'b':[6]}) == 33.3)
    assert(a.merge('[1,2,3]', ['a','b','c']) == ['a','b','c',1,2,3])



def test_argtype_custom():
    class Custom(type):
        def __new__(self, sval):
            return 'x'
    a = ArgType(Custom)
    assert(a.merge('{"a":[1,2,3]}',None) == 'x')

    def custom(sval):
        return 'y'
    a = ArgType(custom)
    assert(a.merge('{"a":[1,2,3]}',None) == 'y')



import pytest
from yaclipy import sub_cmds
from .testutil import exe
from yaclipy.arg_spec import sig_kinds

def nope(*, _input):
    s = f"Nope: {_input}"
    print(s)
    return s


class Bob():
    def __init__(self):
        self.s = ''

    @classmethod
    def classy(cls, y):
        s = f'A Classy function {cls.__name__} {y}'
        print(s)
        return s

    @staticmethod
    def cling(x):
        print(f'cling {x}')
        return f'cling {x}'


    def yes(self, times__t=1, *args):
        self.s += 'w'+'e'*times__t + ' '.join(args)

    @sub_cmds(yes)
    def maybe(self, number__n=3):
        self.s += 'maybe'
        yield 'x'*number__n
        self.s += 'bye'
        return 'zz'

    def mayday(self, number__n=3):
        self.s += 'mayday ' * number__n
        print(f'mayday {self.s}')
        return self

    @sub_cmds(maybe, mayday, nope, cling)
    def call_me(self, my__m='3', z:int=3, **kwargs):
        print(f"---- call_me {id(self)}  {kwargs}  {z}{type(z)}")
        if 'raise' in kwargs: raise RuntimeError(kwargs['raise'])
        self.s += f'{my__m} {z} {kwargs}'
        if 'num' in kwargs:
            return {'self':self, 'number__n':int(kwargs['num'])}



def test_hundred():
    @sub_cmds(Bob.call_me, Bob.mayday)
    def hundred(_a_bob):
        _a_bob.s += 'hundred '
        yield dict(self=_a_bob, number__n=100)
    b = Bob()
    exe(hundred, 'call-me --train tracks', _a_bob=b, zzz=3)
    assert(b.s == "hundred 3 3 {'train': 'tracks', 'number__n': 100}")



def test_vargs_with_defaults():
    def f(pos1=0, pos2=3, *args): pass
    exe(f, '-- and the rest', pos1=1)



def test_bind_array_double_dash():
    def f(a:[int], *, b:[int]) -> list: return a+b
    assert(exe(f, '1 2 3 -b# 4 5 6 -- pop') == 6)




def test_call_func():
    def funky(first__f, *, name__n='sarah', _hidden) -> str:
        return f"FUNKY {first__f}, {name__n} {_hidden}"
    z = exe(funky, 'chair -n high', _hidden=42)
    assert(z == 'FUNKY chair, high 42')
    z = exe(funky, 'chair -n high upper', _hidden=42)
    assert(z == 'FUNKY CHAIR, HIGH 42')



def test_call_input_trumps_default():
    def f(a:int,/,b=3,*,c=4): return (a,b,c)
    assert(exe(f, '1', b=9) == (1,9,4))


def test_call_cmdline_trumps_input():
    def f(a:int,/,b=3,*,c=4): return (a,b,c)
    assert(exe(f, '1 2', b=9) == (1,2,4))



def test_cmd_execute_method_alias():
    def f(my__m='3', **kwargs): return my__m
    assert(exe(f, '') == '3')
    assert(exe(f, 'happy') == 'happy')
    assert(exe(f, 'happy --my__m 99') == '99')



def test_raises():
    with pytest.raises(RuntimeError):
        exe(Bob().call_me, '-m happy --raise bad')
    
    

def test_cmd_subcall():
    bob = Bob()
    r = exe(bob.call_me, '--num 1 mayday')
    assert(bob.s == "3 3 {'num': '1'}mayday ")



def test_cmd_incoming():
    bob = Bob()
    exe(bob.mayday, '4', number__n=2)
    assert(bob.s == 'mayday mayday mayday mayday ')
    bob.s = ''
    exe(bob.mayday, '', number__n=2)
    assert(bob.s == 'mayday mayday ')



def test_cmd_generator():
    s = []
    def get(v=None):
        print(f"Collect {v} + {s}")
        s.append(v)

    @sub_cmds(get)
    def f(times__t=3):
        print(f"f {times__t}")
        for i in range(times__t):
            yield dict(v = i+3)
        return 'x'

    exe(f, '- get')
    assert(s == [3,4,5])
    s[:]=[]
    exe(f, '2 get')
    assert(s == [3,4])



import logging, yaml, time
from print_ext import print
from yaclipy import sub_cmds, Config

log = logging.getLogger('clipy-main')

def run():
    ''' Run the project
    '''
    log.info('Running...')
    log.warn('Uh-oh')
    log.debug('Fixing stuff')



def build():
    ''' Build the project
    '''
    log.info('Building...')
    p = print.progress(steps=42)
    for f in range(42):
        p('Process \b2$', f, '\b$ ...')
        time.sleep(0.1)
    p('Done!', done=True)


    
def deploy():
    ''' Deploy the project
    '''
    log.info('Deploying...')


def configuration():
    ''' Show the current configuration.
    '''
    return Config


@sub_cmds(run, build, deploy, configuration)
def main(*, verbose__v=False, quiet__q=False, config__c=''):
    ''' This is the sole entrypoint for this project.
    
    Parameters:
        --verbose, -v
            Increase the logging level by one notch
        --quiet, -q
            Decrease the logging level by one notch
        --config <option>, -c <option> | default='dev'
            Choose a configuration option
    '''
    level = max(0, min(50, 20 - 10*(int(verbose__v) - int(quiet__q))))
    logging.basicConfig(style='{', format='{levelname:>7} {name:>10} {lineno:<3} | {message}', level=level)
    import config
    Config.configure(config__c or 'dev')



import sys, os, inspect
from print_ext import pretty, Table, Printer, PrettyException, Bdr


class CmdError(PrettyException):
        
    def pretty_help(self):
        cmd = self.cmd
        p = Printer()
        tbl = Table(0.0, 0.0, 1, tmpl='')
        tbl.cell('ALL', cls=Bdr, border=(' ','m:1010'))
        #tbl.cell('R-1', border=('m:1010'))
        #tbl.cell('ALL', cls=Borders, border=(' ','m:1010'))
        tbl.cell('C1', style='1')
        #tbl.cell('R-1', border='m:1110')
        for cmd in sorted(self.cmd.sub_cmds().values(), key=lambda x: x.name):
            tbl('*\t', cmd.name,'\t', pretty(cmd.doc(), fmt='short'),'\t')
        p(tbl)
        if tbl: p.hr(pad=1)
        p.pretty(self.cmd.doc(), pad=1)
        return p


    def __pretty__(self, **kwargs):
        p = self.pretty_help()
        p.hr(self.__class__.__name__, style='err', pad=1)
        for err in self.errors:
            p(' \berr * ', err)
        return p



class CmdHelp(CmdError):
    def __pretty__(self, **kwargs):
        return self.pretty_help()



class CommandNotFound(CmdError): pass



class AmbiguousCommand(CmdError): pass



class CallError(CmdError):
    def __init__(self, cmd):
        self.cmd = cmd
        self.errors = [e[1] for e in cmd.run_spec.errors] if cmd.run_spec else []



class UsageError(PrettyException):
    def __init__(self, fn, *msg):
        super().__init__(msg=msg, fn=fn)

    def __pretty__(self, **kwargs):
        lines, lno = inspect.getsourcelines(self.fn)
        p = Printer(lines[0])
        p('\b2$', os.path.relpath(inspect.getsourcefile(self.fn)), f'\bdem :{lno}')
        return p(*self.msg, pad=1)



#!/usr/bin/env python
import sys
from print_ext import print, PrettyException
import yaclipy as CLI


def donkey(*, _say, name__n="Gordon") -> str:
    ''' Donkey Donkey
    '''
    return f'{name__n} says, \b1 "{_say}"'


def dog(*, _say, times__t=3) -> str:
    ''' Dogs barking
    '''
    return _say + ' bark!'*times__t


@CLI.sub_cmds(dog, donkey)
def main(say, /, *, verbose__v=False):
    ''' Say stuff

    $ ./sub.py "Hello world" -vv donkey

    '''
    if verbose__v: print("\b2 Ready? "*int(verbose__v))
    yield dict(_say=say)
    if verbose__v: print("\bdem Bye")


if __name__ == '__main__':
    try:
        CLI.Command(main)(sys.argv[1:]).run()
    except PrettyException as e:
        print.pretty(e, pad=1)



#!/usr/bin/env python
import sys
from print_ext import print, PrettyException
import yaclipy as CLI

def main(say, times__t=1) -> str:
    ''' Say something multiple times

    Parameters:
        <message>, --say <message>
            What you want to say
        <int>, --times <int>, -t <int>
            How many times you want to say it.
    '''
    return ' '.join([say] * times__t)


if __name__ == '__main__':
    try:
        CLI.Command(main)(sys.argv[1:]).run()
    except PrettyException as e:
        print.pretty(e)



import pytest
from yaclipy.exceptions import *


@pytest.mark.xfail(reason='Not implemented')
def test_uninstall_reqs():
    ''' When requirements are removed from requirements.txt make sure they are uninstalled
    '''
    assert(False)




from print_ext import Printer
from docstring_parser import parse as docstring_parse



class CmdDoc():
    def __init__(self, doc):
        self.doc = docstring_parse(doc)


    def __pretty__(self, fmt='full', **kwargs):
        return getattr(self, f'pretty_{fmt}')()


    def pretty_short(self):
        return self.doc.short_description


    def pretty_full(self):
        p = Printer()
        if self.doc.short_description: p('\bem$',self.doc.short_description, pad=(0,1))
        if self.doc.long_description: p(self.doc.long_description, pad=-1)
        return p



import pytest
from print_ext import print
from yaclipy import Command, sub_cmds
from yaclipy.command import CommandNotFound
from yaclipy.exceptions import *
from .testutil import exe



class Parent():
    def boink(self):
        return 'boink'


class HasCmds(Parent):
    x = 33
    y = lambda x: x+1

    def __init__(self,**kwargs):
        for k,v in kwargs.items(): setattr(self, k,v)

    def __str__(self):
        return ''

    def foo_(self):
        return "There are two foo's"

    def foo(self, x=3):
        return 'foo'*x

    def break_(self, what='everything'):
        return 'break '+what



def test_cmd_retval_sub_cmds():
    def f() -> HasCmds: pass
    subs = Command(f).sub_cmds()
    assert(subs.keys() == {'boink', 'break_', 'foo', 'foo_', 'y'})



def test_cmd_sub_cmds():
    @sub_cmds(str, if_=str, bob_cob=str)
    def f(): pass
    subs = Command(f).sub_cmds()
    assert(subs.keys() == {'str', 'if_', 'bob_cob'})


def test_CommandNotFound():
    @sub_cmds(tst=test_CommandNotFound)
    def f(): pass
    with pytest.raises(CommandNotFound):
        Command(f)(['toss'])



import sys, inspect
from inspect import Parameter
from print_ext import print, Line
from functools import partial
from .arg_spec import ArgSpec, sig_kinds, func_name
from .exceptions import AmbiguousCommand, CommandNotFound, CallError, CmdHelp
from .docs import CmdDoc


def sub_cmds(*args, **kwargs):
    ''' Use this decorator to explicitly set the sub commands for a function.

    If this decorator is not used then the ``dir()`` of the return value annotation is used.
    If there is no return value annotation then no sub commands are available

    The sub commands can be defined as positional parameters, in which case the command name will be the function.__name__.
    If the subcommand is specified as a keyword parameter then the name is used as an alias to the function.
    '''
    def _f(fn):
        fn._sub_cmds = {cmd.real_name:cmd for cmd in [Command(fn) for fn in args]}
        fn._sub_cmds.update({name:Command(fn,name) for name,fn in kwargs.items()})
        return fn
    return _f



class Command():

    def __init__(self, fn, name=None, method=False):
        if name == None: name = func_name(fn)
        self.real_name = name
        self.name = (name[:-1] if name[-1]=='_' else name).replace('_','-')
        self.fn = fn
        self.next_cmd = None
        self.run_spec = None
        self.is_method = method

    def __bool__(self):
        return bool(self.argspec)

    
    @property
    def argspec(self):
        if not hasattr(self, '_argspec'):
            self._argspec = ArgSpec(self.fn, self.is_method)
        return self._argspec
    

    def __call__(self, argv):
        argv = list(argv)
        self.run_spec = self.argspec(argv)
        if '' in self.run_spec.kwargs: raise CmdHelp(cmd=self)
        cmds = self.sub_cmds()
        if argv and not cmds:
            self.run_spec.error('UNUSED', f"Unused trailing parameters: \b1 {argv!r}")
        if self.run_spec.errors: raise CallError(self)
        if not argv: return self
        # Lookup sub-command
        cmd_name = argv[0].replace('_','-')
        cmds = {name:dfn for name, dfn in cmds.items() if dfn.name.startswith(cmd_name)}
        if not cmds:
            raise CommandNotFound(cmd=self, errors = [
                Line(f"Command not found: \b1 {cmd_name}"),
                Line("Valid commands are listed above."),
            ])
        if len(cmds) > 1:
            raise AmbiguousCommand(cmd=self, errors = [
                Line(f"Ambiguous \b1 {cmd_name}\b  matched multiple commands: ", ', '.join(f'\b1 {x}\b ' for x in cmds))
            ])
        self.next_cmd = next(iter(cmds.values()))(argv[1:])
        return self


    def run(self, input=None):
        spec = self.run_spec
        args = list(spec.args)
        kwargs = dict(spec.kwargs)
        sub_self = Parameter.empty
        if hasattr(input, 'items'):
            # Prune input to filter out un-acceptable values
            for k,v in input.items():
                if '**' not in spec.kinds:
                    if k not in spec.params: continue
                    if spec.params[k].kind == Parameter.POSITIONAL_ONLY: continue
                if kwargs.get(k, Parameter.empty) != Parameter.empty: continue
                kwargs[k] = v
        if '_input' in spec.params and spec.params['_input'].kind == Parameter.KEYWORD_ONLY:
            kwargs['_input'] = input
        if spec.has_self: kwargs.setdefault('self', input)
        # Fill in positional empty values
        for p in spec.params.values():         
            if not p.index: break
            if args[p.index-1] != Parameter.empty:
                if p.name in kwargs: del kwargs[p.name]
                continue
            kwargs.setdefault(p.name, p.default)
            if p.name not in kwargs: raise TypeError(f"No value supplied for {p.ordinal} positional parameter '{p.name}'")
            args[p.index-1] = kwargs.pop(p.name)
        # Make the call
        val = self.fn(*args, **kwargs)
        if 'generatorfunction' in spec.kinds:
            return [self._call_next(y) for y in val]
        else:
            return self._call_next(val)
        

    def _call_next(self, val):
        if self.next_cmd == None:
            if val != None: print.pretty(val)
            return val
        return self.next_cmd.run(val)


    def doc(self, check=False):
        if not hasattr(self, '_doc'):
            self._doc = CmdDoc(inspect.getdoc(self.fn) or '')
        return self._doc


    def __pretty__(self, **kwargs):
        p = Printer()
        p(f"Command({self.name!r})")
        p.pretty(self.run_spec or self.argspec)
        for sub in self.sub_cmds().values():
            p(' * ',sub.name)
        return p


    def sub_cmds(self):
        if hasattr(self.fn, '_sub_cmds'): return self.fn._sub_cmds
        retval = self.argspec.retval
        if retval == Parameter.empty: return {}
        dfns = {name:Command(getattr(retval,name), name) for name in dir(retval) if not name.startswith('_')}
        return {name:dfn for name,dfn in dfns.items() if dfn}




import pytest
from print_ext import print
from yaclipy.arg_spec import ArgSpec, ArgType
from yaclipy.exceptions import UsageError


def test_underscore_args():
    def f(help, h, /, cat__c, *, dog__d, _dont__show): pass
    spec = ArgSpec(f)
    print.pretty(spec)
    names = set()
    for p in spec.params.values():
        names.update(p.aliases)
    
    assert(names == {'cat','c','dog','d','help','h'})



def test_hidden_must_be_keyword():
    def f(_ok): pass
    spec = ArgSpec(f)
    def f(_bad,/): pass
    with pytest.raises(UsageError):
        spec = ArgSpec(f)


def test_no_self_aliases():
    def f(_self): pass
    assert(ArgSpec(f).params['_self'].aliases == [])

    def f(self): pass
    assert(ArgSpec(f).params['self'].index == 1)

    def g(bob__self): pass
    with pytest.raises(UsageError):
        spec = ArgSpec(g)

    def h(x__self): pass
    with pytest.raises(UsageError):
        spec = ArgSpec(h)



import traceback, os, copy, inspect, re, functools
from print_ext import print, Table, Text, Printer, PrettyException, pretty


class InvalidOption(PrettyException):
    def __pretty__(self, **kwargs):
        p = Printer()
        p(f'\berr {self.option}\b  is not a valid option.  Valid options are:')
        return p(f'\b1$', '    '.join(sorted(self.options)))



class ConfigOptionUnset(PrettyException): 
    def __pretty__(self, **kwargs):
        p = Printer()
        p(f"You cannot access config variables until \b1 Config.configure()\b  has been called.")
        p(self.tb.line, pad=-1)
        p(os.path.relpath(self.tb.filename), ':', self.tb.lineno)
        return p
    


class SuppressImportError():
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, tb):
        if not type: return
        if len(traceback.extract_tb(tb)) > 1: return
        if type in (ImportError, ModuleNotFoundError): return True
        


class ConfigVar():
    UNSET = object()
    NAME = re.compile(r'\s*(\w+)\s*=')
    
    def __init__(self, config, /, doc='', default=None, coerce=lambda x:x, **user):
        self.dfns = []
        self.doc = doc
        self.config = config
        self.coerce = coerce
        self.default = default
        self.user = user
        self.value = ConfigVar.UNSET
        self._value = ConfigVar.UNSET
        tb = traceback.extract_stack(limit=3)[0]
        name = ConfigVar.NAME.match(tb.line)
        name = name[1] if name else 'undefined'
        self.name = name if tb.name.startswith('<') else f"{tb.name}.{name}"
        self.dfns = [(os.path.relpath(tb.filename), tb.lineno)]


    def reset(self):
        self.set(copy.deepcopy(self.default))
        self.dfns[1:] = []
        

    def set(self, value, offset=0):
        self.value = value
        self._value = ConfigVar.UNSET
        tb = traceback.extract_stack(limit=3+offset)[0]
        self.dfns.append((os.path.relpath(tb.filename), tb.lineno))


    def __call__(self, value=UNSET):
        if self._value == ConfigVar.UNSET and self.value != ConfigVar.UNSET:
            self._value = self.coerce(self.value() if inspect.isroutine(self.value) else self.value)
        if value != ConfigVar.UNSET:
            self.set(value)
        elif self._value == ConfigVar.UNSET:
            tb = traceback.extract_stack(limit=2)[0]
            raise ConfigOptionUnset(tb=tb)
        return self._value
       

    def __lt__(self, other):
        return self.name.lower() < str(other).lower()


    def __str__(self):
        return self.name



class ConfigOption():

    def override(self, name=None): 
        def _override(fn):
            self.fn = functools.partial(fn, self.fn)
            self.set_name(name or fn.__name__)
            return self
        return _override


    def __init__(self, config, /, fn, name=None):
        self.fn = fn
        self.config = config
        self.name = None
        self.set_name(name or fn.__name__)


    def set_name(self, name):
        name = name.replace('_','-')
        if self.name != None and self.name.lower() != name.lower():
            del self.config.options[self.name.lower()]
        self.config.options[name.lower()] = self
        self.name = name


    def __hash__(self):
        return hash(self.name.lower())


    def __eq__(self, other):
        return self.name.lower() == str(other).lower()


    def __repr__(self):
        return self.name


    def __lt__(self, other):
        return self.name.lower() < str(other).lower()



class Configuration():
    def __init__(self):
        self.vars = []
        self.options = {}
        self.cur_option = None
        self.include = SuppressImportError()


    def configure(self, option=''):
        ''' Reconfigure all of the variables with the given `option`.

        Parameters:
            <str>, option=<str> ['']
                The option name to configure or '' to use the defaults
        
        '''
        option = option.lower().replace('_','-')
        if self.cur_option!=None and self.cur_option == option: return
        for v in self.vars: v.reset()
        if not option: return
        if option not in self.options:
            raise InvalidOption(option=option, options=self.options)
        self.options[option].fn()
        self.cur_option = option


    def var(self, *args, **kwargs):
        var = ConfigVar(self, *args, **kwargs)
        self.vars.append(var)
        return var


    def option(self, **kwargs):
        def _f(fn):
            return ConfigOption(self, fn, **kwargs)
        return _f


    def __pretty__(self, **kwargs):
        hide_all = set(name.lower() for name in kwargs.get('hide',[]))
        hide_prefix = set(n[:-1] for n in hide_all if n[-1] == '*')
        def is_hidden(name):
            if name in hide_all: return True
            for prefix in hide_prefix:
                if name.startswith(prefix): return True
        n_hidden = 0
        p = Printer()
        tbl = Table(1,1,2, tmpl='pad')
        tbl.cell('C0', style='1', just='>')
        tbl.cell('C1', style='em')
        for var in sorted(self.vars):
            if is_hidden(var.name.lower()):
                n_hidden += 1
                continue
            tbl(var.name,'\t', pretty(var(),**kwargs),'\t',var.doc,'\t')
        p(tbl, pad=1)
        if n_hidden: p(f'\bwarn {n_hidden} hidden')
        if not self.options: return p
        return p.card('Options\t','\b1$', '    '.join(sorted(self.options)))
        


# A single global configuration
Config = Configuration()



import pytest, traceback
from yaclipy.config import Configuration, ConfigOptionUnset, InvalidOption
from print_ext import pretty
from .testutil import printer, tostr


@pytest.fixture()
def Config():
   return Configuration()



def test_config_var_user(Config):
    v1 = Config.var(mydata=3)
    assert(Config.vars[0].user['mydata'] == 3)



def test_config_var(Config):
    v1 = Config.var()
    v2 = Config.var()
    assert(len(Config.vars) == 2)



def test_config_var_name(Config):
    ab = Config.var()
    a,b = Config.var(), Config.var()
    class Bob():
        jim = Config.var()
    assert(ab.name == 'test_config_var_name.ab')
    assert(a.name == 'test_config_var_name.undefined')
    assert(b.name == 'test_config_var_name.undefined')
    assert(Bob.jim.name == 'Bob.jim')



def test_config_var_coerce(Config):
    v = Config.var('asdf', '123', lambda x: int(x))
    v.reset()
    assert(v() == 123)
    v('asdf')
    with pytest.raises(ValueError):
        v()



def test_config_var_cached(Config):
    x = 1
    def side_effects():
        nonlocal x
        x += 1
        return x
    v = Config.var(default=side_effects)
    v.reset()
    assert(v() == 2)
    assert(v() == 2)
    assert(x == 2)
    v(9)
    assert(v() == 9)
    assert(x == 2)



def test_ConfigOptionUnset(Config):
    v = Config.var(default=3)
    def useit():
        v()
    try:
        useit()
    except ConfigOptionUnset as e:
        v = tostr(pretty(e))
        assert('config_test.py' in v)
        assert('v()' in v)
        return
    assert(False)
    


def test_config_override(Config):
    @Config.option(name='JIM_BO')
    def x(): return 'x'
    assert(Config.options['jim-bo'].fn == x.fn)
    @x.override()
    def Happy_Days(z): return z() == 'x'
    assert('jim-bo' not in Config.options)
    print(Config.options)
    assert(Config.options['happy-days'].fn == Happy_Days.fn)
    assert(Happy_Days.fn())
    @Happy_Days.override(name='zZ_ZZ')
    def asdf(zz): return str(zz())
    assert('happy-days' not in Config.options)
    assert(Config.options['zz-zz'].fn() == 'True')



def test_config_pretty(Config):
    class Bob():
        a = Config.var('desc', 33)
        b = Config.var('d2')
    Config.configure()
    c = tostr(pretty(Config))
    assert(c == '\nBob.a 33   desc\nBob.b None d2\n\n')
    c = tostr(pretty(Config, hide=['Bob.b']))
    assert(c == '\nBob.a 33 desc\n\n1 hidden\n')
    c = tostr(pretty(Config, hide=['Bo*']))
    assert(c == '\n\n2 hidden\n')

    @Config.option()
    def bob(): pass
    @Config.option()
    def d(): pass
    v = tostr(pretty(Config, hide=['Bob.a']))
    assert('d2\n' in v)
    assert('1 hidden' in v)
    assert('Bob.b None' in v)
    assert('bob    d' in v)



def test_config_InvalidOption(Config):
    @Config.option()
    def The_Quick_Fox(): pass
    @Config.option()
    def h(): pass
    
    try:
        Config.configure('jim_Bob')
    except InvalidOption as e:
        v = tostr(pretty(e))
        assert('jim-bob' in v)
        assert('h    the-quick-fox' in v)
        return
    assert(False)



def test_config_include(Config):
    with Config.include: import doesnt.exist
    with Config.include: from . import testutil
    with Config.include: from . import config_import_error
    with pytest.raises(ImportError):
        with Config.include: from .notest import config_import_error
